# mlxtend Machine Learning Library Extensions
#
# A function for plotting a PCA correlation circle
# File Author: Gabriel Azevedo Ferreira <az.fe.gabriel@gmail.com>

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from mlxtend.externals.adjust_text import adjust_text
from mlxtend.feature_extraction import PrincipalComponentAnalysis


def corr2_coeff(A, B):
    """
    Compute correlation coefficients and return as a np array
    """
    A, B = np.array(A), np.array(B)
    # Rowwise mean of input arrays & subtract from input arrays themeselves
    A_mA = A - A.mean(1)[:, None]
    B_mB = B - B.mean(1)[:, None]

    # Sum of squares across rows
    ssA = (A_mA**2).sum(1)
    ssB = (B_mB**2).sum(1)

    # Finally get corr coeff
    return np.dot(A_mA, B_mB.T) / np.sqrt(np.dot(ssA[:, None], ssB[None]))


def create_correlation_table(A, B, names_cols_A, names_cols_B):
    """
    Compute correlation coefficients and return as a DataFrame.

    A and B: 2d array like.
        The columns represent the different variables and the rows are
        the samples of thos variables
    names_cols_A/B : name to be added to the final pandas table

    return: pandas DataFrame with the corelations.Columns and Indexes
    represent the different variables of A and B (respectvely)
    """
    # corrs = np.corrcoef(np.transpose(A), np.transpose(B)
    #                     )[len(names_cols_A):, :len(names_cols_A)]
    corrs = corr2_coeff(A.T, B.T).T

    df_corrs = pd.DataFrame(corrs, columns=names_cols_A, index=names_cols_B)
    return df_corrs


def plot_pca_correlation_graph(
    X,
    variables_names,
    dimensions=(1, 2),
    figure_axis_size=6,
    X_pca=None,
    explained_variance=None,
):
    """
    Compute the PCA for X and plots the Correlation graph

    Parameters
    ----------
    X : 2d array like.
        The columns represent the different variables and the rows are the
         samples of thos variables

    variables_names : array like
        Name of the columns (the variables) of X

    dimensions: tuple with two elements.
        dimensions to be plotted (x,y)

    figure_axis_size :
         size of the final frame. The figure created is a square with length
         and width equal to figure_axis_size.

    X_pca : np.ndarray, shape = [n_samples, n_components].
        Optional.
        `X_pca` is the matrix of the transformed components from X.
        If not provided, the function computes PCA automatically using
        mlxtend.feature_extraction.PrincipalComponentAnalysis
        Expected `n_componentes >= max(dimensions)`

    explained_variance : 1 dimension np.ndarray, length = n_components
        Optional.
        `explained_variance` are the eigenvalues from the diagonalized
        covariance matrix on the PCA transformatiopn.
        If not provided, the function computes PCA independently
        Expected `n_componentes == X.shape[1]`

    Returns
    ----------
        matplotlib_figure, correlation_matrix

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/plotting/plot_pca_correlation_graph/

    """

    X = np.array(X)
    X = X - X.mean(axis=0)
    n_comp = max(dimensions)

    if (X_pca is None) and (explained_variance is None):
        pca = PrincipalComponentAnalysis(n_components=n_comp)
        pca.fit(X)
        X_pca = pca.transform(X)
        explained_variance = pca.e_vals_

    elif (X_pca is not None) and (explained_variance is None):
        raise ValueError(
            "If `X_pca` is not None, the `explained variance`"
            " values should not be `None`."
        )

    elif (X_pca is None) and (explained_variance is not None):
        raise ValueError(
            "If `explained variance` is not None, the `X_pca`"
            " values should not be `None`."
        )

    elif (X_pca is not None) and (explained_variance is not None):
        if X_pca.shape[1] != len(explained_variance):
            raise ValueError(
                f"Number of principal components must "
                f"match the number "
                f"of eigenvalues. Got "
                f"{X_pca.shape[1]} "
                f"!= "
                f"{len(explained_variance)}"
            )

    if X_pca.shape[1] < n_comp:
        raise ValueError(
            f"Input array `X_pca` contains fewer principal"
            f" components than expected based on `dimensions`."
            f" Got {X_pca.shape[1]} components in X_pca, expected"
            f" at least `max(dimensions)={n_comp}`."
        )
    if len(explained_variance) < n_comp:
        raise ValueError(
            f"Input array `explained_variance` contains fewer"
            f" elements than expected. Got"
            f" {len(explained_variance)} elements, expected"
            f"`X.shape[1]={X.shape[1]}`."
        )

    corrs = create_correlation_table(
        X_pca, X, ["Dim " + str(i + 1) for i in range(n_comp)], variables_names
    )
    tot = sum(X.var(0)) * X.shape[0] / (X.shape[0] - 1)
    explained_var_ratio = [(i / tot) * 100 for i in explained_variance]

    # Plotting circle
    fig_res = plt.figure(figsize=(figure_axis_size, figure_axis_size))
    plt.Circle((0, 0), radius=1, color="k", fill=False)
    circle1 = plt.Circle((0, 0), radius=1, color="k", fill=False)
    fig = plt.gcf()
    fig.gca().add_artist(circle1)

    # Plotting arrows
    texts = []
    for name, row in corrs.iterrows():
        x = row["Dim " + str(dimensions[0])]
        y = row["Dim " + str(dimensions[1])]
        plt.arrow(0.0, 0.0, x, y, color="k", length_includes_head=True, head_width=0.05)

        plt.plot([0.0, x], [0.0, y], "k-")
        texts.append(plt.text(x, y, name, fontsize=2 * figure_axis_size))
    # Plotting vertical lines
    plt.plot([-1.1, 1.1], [0, 0], "k--")
    plt.plot([0, 0], [-1.1, 1.1], "k--")

    # Adjusting text
    adjust_text(texts)
    # Setting limits and title
    plt.xlim((-1.1, 1.1))
    plt.ylim((-1.1, 1.1))
    plt.title("Correlation Circle", fontsize=figure_axis_size * 3)

    plt.xlabel(
        "Dim "
        + str(dimensions[0])
        + " (%s%%)" % str(explained_var_ratio[dimensions[0] - 1])[:4],
        fontsize=figure_axis_size * 2,
    )
    plt.ylabel(
        "Dim "
        + str(dimensions[1])
        + " (%s%%)" % str(explained_var_ratio[dimensions[1] - 1])[:4],
        fontsize=figure_axis_size * 2,
    )
    return fig_res, corrs



# mlxtend Machine Learning Library Extensions
#
# Time series cross validation with grouping.
# Author: Dmitry Labazkin <labdmitriy@gmail.com>
#
# License: BSD 3 clause

from itertools import accumulate, chain, groupby, islice

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Patch
from matplotlib.ticker import MaxNLocator
from sklearn.utils import indexable


class GroupTimeSeriesSplit:
    """Group time series cross-validator.

    Parameters
    ----------
    test_size : int
        Size of test dataset.
    train_size : int (default=None)
        Size of train dataset.
    n_splits : int (default=None)
        Number of the splits.
    gap_size : int (default=0)
        Gap size between train and test datasets.
    shift_size : int (default=1)
        Step to shift for the next fold.
    window_type : str (default="rolling")
        Type of the window. Possible values: "rolling", "expanding".

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/GroupTimeSeriesSplit/
    """

    def __init__(
        self,
        test_size,
        train_size=None,
        n_splits=None,
        gap_size=0,
        shift_size=1,
        window_type="rolling",
    ):
        if (train_size is None) and (n_splits is None):
            raise ValueError("Either train_size or n_splits should be defined")

        if window_type not in ["rolling", "expanding"]:
            raise ValueError('Window type can be either "rolling" or "expanding"')

        if (train_size is not None) and (window_type == "expanding"):
            raise ValueError("Train size can be specified only with rolling window")

        self.test_size = test_size
        self.train_size = train_size
        self.n_splits = n_splits
        self.gap_size = gap_size
        self.shift_size = shift_size
        self.window_type = window_type

    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like
            Training data.
        y : array-like (default=None)
            Always ignored, exists for compatibility.
        groups : array-like (default=None)
            Array with group names or sequence numbers.

        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """
        test_size = self.test_size
        gap = self.gap_size
        shift_size = self.shift_size
        X, y, groups = indexable(X, y, groups)

        if groups is None:
            raise ValueError("The groups should be specified")

        group_names, group_lengths = zip(
            *[
                (group_name, len(list(group_seq)))
                for group_name, group_seq in groupby(groups)
            ]
        )
        n_groups = len(group_names)

        if n_groups != len(set(group_names)):
            raise ValueError("The groups should be consecutive")

        self._n_groups = n_groups
        group_starts_idx = chain(
            [0],
            islice(accumulate(group_lengths), len(group_lengths) - 1),
        )
        groups_dict = dict(zip(group_names, group_starts_idx))
        n_samples = len(X)

        self._calculate_split_params()
        train_size = self.train_size
        n_splits = self.n_splits
        train_start_idx = self._train_start_idx
        train_end_idx = train_start_idx + train_size
        test_start_idx = train_end_idx + gap
        test_end_idx = test_start_idx + test_size

        for _ in range(n_splits):
            train_idx = np.r_[
                slice(
                    groups_dict[group_names[train_start_idx]],
                    groups_dict[group_names[train_end_idx]],
                )
            ]

            if test_end_idx < n_groups:
                test_idx = np.r_[
                    slice(
                        groups_dict[group_names[test_start_idx]],
                        groups_dict[group_names[test_end_idx]],
                    )
                ]
            else:
                test_idx = np.r_[
                    slice(groups_dict[group_names[test_start_idx]], n_samples)
                ]

            yield train_idx, test_idx

            if self.window_type == "rolling":
                train_start_idx = train_start_idx + shift_size

            train_end_idx = train_end_idx + shift_size
            test_start_idx = test_start_idx + shift_size
            test_end_idx = test_end_idx + shift_size

    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations in the cross-validator.

        Parameters
        ----------
        X : object
            Always ignored, exists for compatibility.
        y : object
            Always ignored, exists for compatibility.
        groups : object
            Always ignored, exists for compatibility.

        Returns
        -------
        n_splits : int
            Returns the number of splitting iterations in the cross-validator.
        """
        return self.n_splits

    def _calculate_split_params(self):
        train_size = self.train_size
        test_size = self.test_size
        n_splits = self.n_splits
        gap = self.gap_size
        shift_size = self.shift_size
        n_groups = self._n_groups

        not_enough_data_error = (
            "Not enough data to split number of groups ({0})"
            " for number splits ({1})"
            " with train size ({2}), test size ({3}),"
            " gap size ({4}), shift size ({5})"
        )

        if (train_size is None) and (n_splits is not None):
            train_size = n_groups - gap - test_size - (n_splits - 1) * shift_size
            self.train_size = train_size

            if train_size <= 0:
                raise ValueError(
                    not_enough_data_error.format(
                        n_groups,
                        n_splits,
                        train_size,
                        test_size,
                        gap,
                        shift_size,
                    )
                )
            train_start_idx = 0
        elif (n_splits is None) and (train_size is not None):
            n_splits = (n_groups - train_size - gap - test_size) // shift_size + 1
            self.n_splits = n_splits

            if self.n_splits <= 0:
                raise ValueError(
                    not_enough_data_error.format(
                        n_groups,
                        n_splits,
                        train_size,
                        test_size,
                        gap,
                        shift_size,
                    )
                )
            train_start_idx = (
                n_groups - train_size - gap - test_size - (n_splits - 1) * shift_size
            )
        else:
            train_start_idx = (
                n_groups - train_size - gap - test_size - (n_splits - 1) * shift_size
            )

            if train_start_idx < 0:
                raise ValueError(
                    not_enough_data_error.format(
                        n_groups,
                        n_splits,
                        train_size,
                        test_size,
                        gap,
                        shift_size,
                    )
                )

        self._train_start_idx = train_start_idx


def print_split_info(X, y, groups, **cv_args):
    """Print information details about splits."""
    cv = GroupTimeSeriesSplit(**cv_args)
    groups = np.array(groups)

    for train_idx, test_idx in cv.split(X, groups=groups):
        print("Train indices:", train_idx)
        print("Test indices:", test_idx)
        print("Train length:", len(train_idx))
        print("Test length:", len(test_idx))
        print("Train groups:", groups[train_idx])
        print("Test groups:", groups[test_idx])
        print("Train group size:", len(set(groups[train_idx])))
        print("Test group size:", len(set(groups[test_idx])))
        print("Train group months:", X.index[train_idx].values)
        print("Test group months:", X.index[test_idx].values)
        print()


def plot_split_indices(cv, cv_args, X, y, groups, n_splits, image_file_path=None):
    """Create a sample plot for indices of a cross-validation object."""
    fig, ax = plt.subplots(figsize=(12, 4))
    cmap_data = plt.cm.tab20
    cmap_cv = plt.cm.coolwarm
    lw = 10
    marker_size = 200

    for split_idx, (train_idx, test_idx) in enumerate(
        cv.split(X=X, y=y, groups=groups)
    ):
        indices = np.array([np.nan] * len(X))
        indices[test_idx] = 1
        indices[train_idx] = 0

        ax.scatter(
            range(len(X)),
            [split_idx + 0.5] * len(X),
            c=indices,
            marker="_",
            lw=lw,
            cmap=cmap_cv,
            vmin=-0.4,
            vmax=1.4,
            s=marker_size,
        )

    ax.scatter(
        range(len(X)),
        [split_idx + 1.5] * len(X),
        c=groups,
        marker="_",
        lw=lw,
        cmap=cmap_data,
        s=marker_size,
    )

    yticklabels = list(range(n_splits)) + ["group"]
    ax.set(
        yticks=np.arange(n_splits + 1) + 0.5,
        yticklabels=yticklabels,
        ylabel="CV iteration",
        ylim=[n_splits + 1.2, -0.2],
        xlim=[-0.5, len(indices) - 0.5],
    )

    ax.legend(
        [Patch(color=cmap_cv(0.2)), Patch(color=cmap_cv(0.8))],
        ["Training set", "Testing set"],
        loc=(1.02, 0.8),
        fontsize=13,
    )

    ax.set_title("{}\n{}".format(type(cv).__name__, cv_args), fontsize=15)
    ax.xaxis.set_major_locator(MaxNLocator(min_n_ticks=len(X), integer=True))
    ax.set_xlabel(xlabel="Sample index", fontsize=13)
    ax.set_ylabel(ylabel="CV iteration", fontsize=13)
    ax.tick_params(axis="both", which="major", labelsize=13)
    ax.tick_params(axis="both", which="minor", labelsize=13)

    plt.tight_layout()

    if image_file_path:
        plt.savefig(image_file_path, bbox_inches="tight")

    plt.show()


def plot_splits(X, y, groups, image_file_path=None, **cv_args):
    """Visualize splits by group."""
    cv = GroupTimeSeriesSplit(**cv_args)
    cv._n_groups = len(np.unique(groups))
    cv._calculate_split_params()
    n_splits = cv.n_splits

    plot_split_indices(
        cv, cv_args, X, y, groups, n_splits, image_file_path=image_file_path
    )


def print_cv_info(cv, X, y, groups, clf, scores):
    """Print information details about cross-validation usage with classifier."""
    for split_idx, (train_idx, test_idx) in enumerate(cv.split(X, y, groups)):
        clf.fit(X.iloc[train_idx], y.iloc[train_idx])
        y_train_pred = clf.predict(X.iloc[train_idx])
        y_test_pred = clf.predict(X.iloc[test_idx])
        print(f"Split number: {split_idx + 1}")
        print(f"Train true target: {y.iloc[train_idx].values}")
        print(f"Train predicted target: {y_train_pred}")
        print(f"Test true target: {y.iloc[test_idx].values}")
        print(f"Test predicted target: {y_test_pred}")
        print(f"Accuracy: {scores[split_idx].round(2)}")
        print()



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal

from mlxtend.data import iris_data
from mlxtend.feature_extraction import PrincipalComponentAnalysis as PCA
from mlxtend.preprocessing import standardize
from mlxtend.utils import assert_raises

X, y = iris_data()
X_std = standardize(X)


def test_default_components():
    pca = PCA()
    res = pca.fit(X_std).transform(X_std)
    assert res.shape[1] == 4


def test_whitening():
    pca = PCA(n_components=2)
    res = pca.fit(X_std).transform(X_std)
    diagonals_sum = np.sum(np.diagonal(np.cov(res.T)))
    assert round(diagonals_sum, 1) == 3.9, diagonals_sum

    pca = PCA(n_components=2, whitening=True)
    res = pca.fit(X_std).transform(X_std)
    diagonals_sum = np.sum(np.diagonal(np.cov(res.T)))
    assert round(diagonals_sum, 1) == 2.0, diagonals_sum


def test_default_2components():
    pca = PCA(n_components=2)
    res = pca.fit(X_std).transform(X_std)
    assert res.shape[1] == 2


def test_eigen_vs_svd():
    pca = PCA(n_components=2, solver="eigen")
    eigen_res = pca.fit(X_std).transform(X_std)
    pca = PCA(n_components=2, solver="svd")
    svd_res = pca.fit(X_std).transform(X_std)
    assert_allclose(np.absolute(eigen_res), np.absolute(svd_res), atol=0.0001)


def test_default_components_zero():
    assert_raises(AttributeError, "n_components must be > 1 or None", PCA, 0)


def test_evals():
    pca = PCA(n_components=2, solver="eigen")
    pca.fit(X_std)

    expected = [2.93035378, 0.92740362, 0.14834223, 0.02074601]
    assert_almost_equal(pca.e_vals_, expected, decimal=5)

    pca = PCA(n_components=2, solver="svd")
    pca.fit(X_std)
    assert_almost_equal(pca.e_vals_, expected, decimal=5)


def test_loadings():
    expect = np.array(
        [
            [0.9, -0.4, -0.3, 0.0],
            [-0.5, -0.9, 0.1, -0.0],
            [1.0, -0.0, 0.1, -0.1],
            [1.0, -0.1, 0.2, 0.1],
        ]
    )

    pca = PCA(solver="eigen")
    pca.fit(X_std)
    assert_almost_equal(pca.loadings_, expect, decimal=1)

    expect = np.array(
        [
            [-0.9, -0.4, 0.3, 0.0],
            [0.4, -0.9, -0.1, -0.0],
            [-1.0, -0.0, -0.1, -0.1],
            [-1.0, -0.1, -0.2, 0.1],
        ]
    )

    pca = PCA(solver="svd")
    pca.fit(X_std)
    assert_almost_equal(pca.loadings_, expect, decimal=1)


def test_fail_array_dimension():
    pca = PCA(n_components=2)
    assert_raises(
        ValueError, "X must be a 2D array. Try X[:, numpy.newaxis]", pca.fit, X_std[1]
    )


def test_fail_array_dimension_2():
    pca = PCA(n_components=2)
    assert_raises(
        ValueError,
        "X must be a 2D array. Try X[:, numpy.newaxis]",
        pca.transform,
        X_std[1],
    )


def test_variance_explained_ratio():
    pca = PCA()
    pca.fit(X_std)
    assert_almost_equal(np.sum(pca.e_vals_normalized_), 1.0)
    assert np.sum(pca.e_vals_normalized_ < 0.0) == 0


def test_pca_on_uncentered_data():
    pca1 = PCA(solver="svd")
    pca1.fit(X)

    pca2 = PCA(solver="eigen")
    pca2.fit(X)
    assert_almost_equal(pca1.e_vals_normalized_, pca2.e_vals_normalized_)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np

from mlxtend.classifier import SoftmaxRegression
from mlxtend.data import iris_data
from mlxtend.plotting import plot_decision_regions
from mlxtend.utils import assert_raises

plt.switch_backend("agg")
X, y = iris_data()
sr = SoftmaxRegression(epochs=15, eta=0.01, random_seed=1)


def test_pass():
    sr.fit(X[:, :2], y)
    plot_decision_regions(X=X[:, :2], y=y, clf=sr)


def test_ylist():
    sr.fit(X[:, :2], y)
    assert_raises(
        ValueError,
        "y must be a NumPy array. Found {}".format(type([])),
        plot_decision_regions,
        X[:, :2],
        list(y),
        sr,
    )


def test_filler_feature_values_fail():
    sr.fit(X, y)
    assert_raises(
        ValueError,
        "Filler values must be provided when " "X has more than 2 training features.",
        plot_decision_regions,
        X,
        y,
        sr,
    )


def test_feature_index_fail():
    sr.fit(X, y)
    assert_raises(
        ValueError,
        "Unable to unpack feature_index. Make sure feature_index "
        "only has two dimensions.",
        plot_decision_regions,
        X,
        y,
        sr,
        feature_index=(0, 1, 2),
        filler_feature_values={2: 0.5},
    )


def test_X_dim_fail():
    sr.fit(X[:, :2], y)
    assert_raises(
        ValueError,
        "X must be a 2D array",
        plot_decision_regions,
        X[:, :2].flatten(),
        y,
        sr,
    )


def test_X_highlight_fail_if_1d():
    sr.fit(X[:, :2], y)
    assert_raises(
        ValueError,
        "X_highlight must be a 2D array",
        plot_decision_regions,
        X[:, :2],
        y,
        sr,
        X_highlight=y,
    )


def test_y_int_ary():
    sr.fit(X[:, :2], y)
    assert_raises(
        ValueError,
        "y must be an integer array. Found float64. "
        "Try passing the array as y.astype(np.int_)",
        plot_decision_regions,
        X[:, :2],
        y.astype(np.float_),
        sr,
    )


def test_y_ary_dim():
    sr.fit(X[:, :2], y)
    assert_raises(
        ValueError,
        "y must be a 1D array",
        plot_decision_regions,
        X[:, :2],
        y[:, np.newaxis],
        sr,
    )


def test_scatter_kwargs_type():
    kwargs = "not a dictionary"
    sr.fit(X[:, :2], y)
    message = "d must be of type dict or None, but got " "{} instead".format(
        type(kwargs)
    )
    assert_raises(
        TypeError,
        message,
        plot_decision_regions,
        X[:, :2],
        y,
        sr,
        scatter_kwargs=kwargs,
    )


def test_contourf_kwargs_type():
    kwargs = "not a dictionary"
    sr.fit(X[:, :2], y)
    message = "d must be of type dict or None, but got " "{} instead".format(
        type(kwargs)
    )
    assert_raises(
        TypeError,
        message,
        plot_decision_regions,
        X[:, :2],
        y,
        sr,
        contourf_kwargs=kwargs,
    )


def test_scatter_highlight_kwargs_type():
    kwargs = "not a dictionary"
    sr.fit(X[:, :2], y)
    message = "d must be of type dict or None, but got " "{} instead".format(
        type(kwargs)
    )
    assert_raises(
        TypeError,
        message,
        plot_decision_regions,
        X[:, :2],
        y,
        sr,
        X_highlight=X[:, :2],
        scatter_highlight_kwargs=kwargs,
    )



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend.evaluate import mcnemar_tables
from mlxtend.utils import assert_raises


def test_input_array_1d():
    t = np.array([[1, 2], [3, 4]])
    assert_raises(
        ValueError,
        "One or more input arrays are not 1-dimensional.",
        mcnemar_tables,
        t,
        t,
        t,
    )


def test_input_array_lengths_1():
    t = np.array([1, 2])
    t2 = np.array([1, 2, 3])
    assert_raises(
        ValueError,
        ("Each prediction array must have" " the same number of samples."),
        mcnemar_tables,
        t,
        t2,
        t,
    )


def test_model_have_same_len():
    y_true = np.array([1, 1, 0])
    y_1 = np.array([0, 1, 0])
    y_2 = np.array([1, 1, 0])
    y_3 = np.array([1, 1])

    assert_raises(
        ValueError,
        ("Each prediction array must have" " the same number of samples."),
        mcnemar_tables,
        y_true,
        y_1,
        y_2,
        y_3,
    )


def test_min_number_of_models():
    y_true = np.array([1, 1, 0])
    y_1 = np.array([0, 1, 0])

    assert_raises(
        ValueError,
        ("Provide at least 2 model prediction arrays."),
        mcnemar_tables,
        y_true,
        y_1,
    )


def test_input_binary_all_right():
    y_target = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    y_model1 = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    y_model2 = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    tb = mcnemar_tables(y_target, y_model1, y_model2)
    expect = np.array([[8, 0], [0, 0]])

    assert len(tb) == 1
    np.testing.assert_array_equal(tb["model_0 vs model_1"], expect)


def test_input_binary_all_wrong():
    y_target = np.array([1, 1, 1, 1, 0, 0, 0, 0])
    y_model1 = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    y_model2 = np.array([0, 0, 0, 0, 1, 1, 1, 1])
    tb = mcnemar_tables(y_target, y_model1, y_model2)
    expect = np.array([[0, 0], [0, 8]])

    assert len(tb) == 1
    np.testing.assert_array_equal(tb["model_0 vs model_1"], expect)


def test_input_binary():
    y_target = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
    y_model1 = np.array([0, 1, 0, 0, 0, 1, 1, 0, 0, 0])
    y_model2 = np.array([0, 0, 1, 1, 0, 1, 1, 0, 0, 0])
    y_model3 = np.array([0, 1, 1, 0, 0, 1, 1, 0, 0, 0])
    tb = mcnemar_tables(y_target, y_model1, y_model2, y_model3)

    expect1 = np.array([[4, 2], [1, 3]])

    expect2 = np.array([[5, 1], [0, 4]])

    expect3 = np.array([[4, 1], [1, 4]])

    assert len(tb) == 3
    np.testing.assert_array_equal(tb["model_0 vs model_1"], expect1)
    np.testing.assert_array_equal(tb["model_0 vs model_2"], expect2)
    np.testing.assert_array_equal(tb["model_1 vs model_2"], expect3)


def test_input_nonbinary():
    y_target = np.array([0, 0, 0, 0, 0, 2, 2, 3, 3, 3])
    y_model1 = np.array([0, 1, 0, 0, 0, 2, 2, 1, 1, 1])
    y_model2 = np.array([0, 0, 1, 1, 0, 2, 2, 0, 1, 0])
    y_model3 = np.array([0, 1, 1, 0, 0, 2, 2, 0, 1, 1])
    tb = mcnemar_tables(y_target, y_model1, y_model2, y_model3)

    expect1 = np.array([[4, 2], [1, 3]])

    expect2 = np.array([[5, 1], [0, 4]])

    expect3 = np.array([[4, 1], [1, 4]])

    assert len(tb) == 3
    np.testing.assert_array_equal(tb["model_0 vs model_1"], expect1)
    np.testing.assert_array_equal(tb["model_0 vs model_2"], expect2)
    np.testing.assert_array_equal(tb["model_1 vs model_2"], expect3)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from sklearn.base import clone

from mlxtend.classifier import MultiLayerPerceptron as MLP
from mlxtend.data import iris_data
from mlxtend.utils import assert_raises

X, y = iris_data()
X = X[:, [0, 3]]  # sepal length and petal width
X_bin = X[0:100]  # class 0 and class 1
y_bin = y[0:100]  # class 0 and class 1

# standardize
X_bin[:, 0] = (X_bin[:, 0] - X_bin[:, 0].mean()) / X_bin[:, 0].std()
X_bin[:, 1] = (X_bin[:, 1] - X_bin[:, 1].mean()) / X_bin[:, 1].std()
X[:, 0] = (X[:, 0] - X[:, 0].mean()) / X[:, 0].std()
X[:, 1] = (X[:, 1] - X[:, 1].mean()) / X[:, 1].std()


def test_multiclass_gd_acc():
    mlp = MLP(epochs=20, eta=0.05, hidden_layers=[10], minibatches=1, random_seed=1)
    mlp.fit(X, y)
    assert round(mlp.cost_[0], 2) == 0.55, mlp.cost_[0]
    assert round(mlp.cost_[-1], 2) == 0.01, mlp.cost_[-1]
    assert (y == mlp.predict(X)).all()


def test_progress_1():
    mlp = MLP(
        epochs=1,
        eta=0.05,
        hidden_layers=[10],
        minibatches=1,
        print_progress=1,
        random_seed=1,
    )
    mlp.fit(X, y)


def test_progress_2():
    mlp = MLP(
        epochs=1,
        eta=0.05,
        hidden_layers=[10],
        minibatches=1,
        print_progress=2,
        random_seed=1,
    )
    mlp.fit(X, y)


def test_progress_3():
    mlp = MLP(
        epochs=1,
        eta=0.05,
        hidden_layers=[10],
        minibatches=1,
        print_progress=3,
        random_seed=1,
    )
    mlp.fit(X, y)


def test_predict_proba():
    mlp = MLP(epochs=20, eta=0.05, hidden_layers=[10], minibatches=1, random_seed=1)
    mlp.fit(X, y)

    pred = mlp.predict_proba(X[0, np.newaxis])
    exp = np.array([[0.6, 0.2, 0.2]])
    np.testing.assert_almost_equal(pred, exp, decimal=1)


def test_multiclass_sgd_acc():
    mlp = MLP(
        epochs=20, eta=0.05, hidden_layers=[25], minibatches=len(y), random_seed=1
    )
    mlp.fit(X, y)
    assert round(mlp.cost_[-1], 3) == 0.023, mlp.cost_[-1]
    assert (y == mlp.predict(X)).all()


def test_multiclass_minibatch_acc():
    mlp = MLP(epochs=20, eta=0.05, hidden_layers=[25], minibatches=5, random_seed=1)
    mlp.fit(X, y)
    assert round(mlp.cost_[-1], 3) == 0.024, mlp.cost_[-1]
    assert (y == mlp.predict(X)).all()


def test_num_hidden_layers():
    assert_raises(
        AttributeError,
        "Currently, only 1 hidden layer is supported",
        MLP,
        20,
        0.05,
        [25, 10],
    )


def test_binary_gd():
    mlp = MLP(epochs=20, eta=0.05, hidden_layers=[25], minibatches=5, random_seed=1)

    mlp.fit(X_bin, y_bin)
    assert (y_bin == mlp.predict(X_bin)).all()


def test_score_function():
    mlp = MLP(epochs=20, eta=0.05, hidden_layers=[25], minibatches=5, random_seed=1)
    mlp.fit(X, y)
    acc = mlp.score(X, y)
    assert acc == 1.0, acc


def test_decay_function():
    mlp = MLP(
        epochs=20,
        eta=0.05,
        decrease_const=0.01,
        hidden_layers=[25],
        minibatches=5,
        random_seed=1,
    )

    mlp.fit(X, y)
    assert mlp._decr_eta < mlp.eta
    acc = mlp.score(X, y)
    assert round(acc, 2) == 0.98, acc


def test_momentum_1():
    mlp = MLP(
        epochs=20,
        eta=0.05,
        momentum=0.1,
        hidden_layers=[25],
        minibatches=len(y),
        random_seed=1,
    )

    mlp.fit(X, y)
    assert round(mlp.cost_[-1], 4) == 0.0057, mlp.cost_[-1]
    assert (y == mlp.predict(X)).all()


def test_retrain():
    mlp = MLP(epochs=5, eta=0.05, hidden_layers=[10], minibatches=len(y), random_seed=1)

    mlp.fit(X, y)
    cost_1 = mlp.cost_[-1]
    mlp.fit(X, y)
    cost_2 = mlp.cost_[-1]
    mlp.fit(X, y, init_params=False)
    cost_3 = mlp.cost_[-1]

    assert cost_2 == cost_1
    assert cost_3 < (cost_2 / 2.0)


def test_clone():
    mlp = MLP(epochs=5, eta=0.05, hidden_layers=[10], minibatches=len(y), random_seed=1)

    clone(mlp)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Functions for tokenizing text data.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import re


def tokenizer_words_and_emoticons(text):
    """Convert text to lowercase words and emoticons.

    Examples
    -----------
    >>> tokenizer_words_and_emoticons('</a>This :) is :( a test :-)!')
    ['this', 'is', 'a', 'test', ':)', ':(', ':-)']

    For more usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/text/tokenizer_words_and_emoticons/

    """
    text = re.sub(r"<[^>]*>", "", text)
    emoticons = re.findall(r"(?::|;|=)(?:-)?(?:\)|\(|D|P)", text)
    text = re.sub(r"[\W]+", " ", text.lower()) + " ".join(emoticons)
    return text.split()


def tokenizer_emoticons(text):
    """Return emoticons from text

    Examples
    -----------
    >>> tokenizer_emoticons('</a>This :) is :( a test :-)!')
    [':)', ':(', ':-)']

    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/text/tokenizer_emoticons/

    """
    text = re.sub(r"<[^>]*>", "", text)
    emoticons = re.findall(r"(?::|;|=)(?:-)?(?:\)|\(|D|P)", text)
    return emoticons



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend._base import _IterativeModel


class BlankModel(_IterativeModel):
    def __init__(self, print_progress=0, random_seed=1):
        self.print_progress = print_progress
        self.random_seed = random_seed
        np.random.seed(random_seed)


def test_init():
    est = BlankModel(print_progress=0, random_seed=1)
    assert hasattr(est, "print_progress")
    assert hasattr(est, "random_seed")


def test_shuffle():
    X = np.array([[1], [2], [3]])
    y = np.array([1, 2, 3])
    est = BlankModel(print_progress=0, random_seed=1)
    X_sh, y_sh = est._shuffle_arrays(arrays=[X, np.array(y)])
    np.testing.assert_equal(X_sh, np.array([[1], [3], [2]]))
    np.testing.assert_equal(y_sh, np.array([1, 3, 2]))


def test_init_params():
    est = BlankModel(print_progress=0, random_seed=1)
    b, w = est._init_params(weights_shape=(3, 3), bias_shape=(1,), random_seed=0)
    assert b == np.array([0.0]), b

    expect_w = np.array(
        [[0.018, 0.004, 0.01], [0.022, 0.019, -0.01], [0.01, -0.002, -0.001]]
    )
    np.testing.assert_almost_equal(w, expect_w, decimal=3)


def test_minibatches_divisible():
    ary = np.array([1, 2, 3, 4, 5, 6, 7, 8])
    est = BlankModel(print_progress=0, random_seed=1)
    rgen = np.random.RandomState(1)
    gen_arys = est._yield_minibatches_idx(rgen=rgen, n_batches=2, data_ary=ary)
    arys = list(gen_arys)

    assert (arys[0] == np.array([7, 2, 1, 6])).all()
    assert (arys[1] == np.array([0, 4, 3, 5])).all()


def test_minibatches_remainder():
    ary = np.array([1, 2, 3, 4, 5, 6, 7])
    est = BlankModel(print_progress=0, random_seed=1)
    rgen = np.random.RandomState(1)
    gen_arys = est._yield_minibatches_idx(rgen=rgen, n_batches=2, data_ary=ary)
    arys = list(gen_arys)

    assert len(arys) == 2
    assert (arys[0] == np.array([6, 2, 1])).all()
    assert (arys[1] == np.array([0, 4, 3, 5])).all()


def test_minibatch_1sample():
    ary = np.array([1, 2, 3, 4, 5, 6, 7])
    est = BlankModel(print_progress=0, random_seed=1)
    rgen = np.random.RandomState(1)
    gen_arys = est._yield_minibatches_idx(rgen=rgen, n_batches=7, data_ary=ary)
    arys = list(gen_arys)

    assert len(arys) == 7
    assert arys[0] == np.array([6]), arys[0]


def test_minibatch_allsample():
    ary = np.array([1, 2, 3, 4, 5, 6, 7])
    est = BlankModel(print_progress=0, random_seed=1)
    rgen = np.random.RandomState(1)
    gen_arys = est._yield_minibatches_idx(
        rgen=rgen, n_batches=1, data_ary=ary, shuffle=False
    )
    arys = list(gen_arys)[0]
    assert (arys == np.array([0, 1, 2, 3, 4, 5, 6])).all()



import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_raises as numpy_assert_raises

from mlxtend.frequent_patterns import apriori, association_rules

one_ary = np.array(
    [
        [0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 1],
        [0, 0, 1, 1, 0, 1, 0, 1, 1, 0, 1],
        [1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1],
        [0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0],
    ]
)

cols = [
    "Apple",
    "Corn",
    "Dill",
    "Eggs",
    "Ice cream",
    "Kidney Beans",
    "Milk",
    "Nutmeg",
    "Onion",
    "Unicorn",
    "Yogurt",
]

df = pd.DataFrame(one_ary, columns=cols)

df_freq_items = apriori(df, min_support=0.6)

df_freq_items_with_colnames = apriori(df, min_support=0.6, use_colnames=True)

columns_ordered = [
    "antecedents",
    "consequents",
    "antecedent support",
    "consequent support",
    "support",
    "confidence",
    "lift",
    "leverage",
    "conviction",
    "zhangs_metric",
]


def test_default():
    res_df = association_rules(df_freq_items)
    res_df["antecedents"] = res_df["antecedents"].apply(lambda x: str(frozenset(x)))
    res_df["consequents"] = res_df["consequents"].apply(lambda x: str(frozenset(x)))
    res_df.sort_values(columns_ordered, inplace=True)
    res_df.reset_index(inplace=True, drop=True)

    expect = pd.DataFrame(
        [
            [(8,), (5,), 0.6, 1.0, 0.6, 1.0, 1.0, 0.0, np.inf, 0],
            [(6,), (5,), 0.6, 1.0, 0.6, 1.0, 1.0, 0.0, np.inf, 0],
            [(8, 3), (5,), 0.6, 1.0, 0.6, 1.0, 1.0, 0.0, np.inf, 0],
            [(8, 5), (3,), 0.6, 0.8, 0.6, 1.0, 1.25, 0.12, np.inf, 0.5],
            [(8,), (3, 5), 0.6, 0.8, 0.6, 1.0, 1.25, 0.12, np.inf, 0.5],
            [(3,), (5,), 0.8, 1.0, 0.8, 1.0, 1.0, 0.0, np.inf, 0],
            [(5,), (3,), 1.0, 0.8, 0.8, 0.8, 1.0, 0.0, 1.0, 0],
            [(10,), (5,), 0.6, 1.0, 0.6, 1.0, 1.0, 0.0, np.inf, 0],
            [(8,), (3,), 0.6, 0.8, 0.6, 1.0, 1.25, 0.12, np.inf, 0.5],
        ],
        columns=columns_ordered,
    )

    expect["antecedents"] = expect["antecedents"].apply(lambda x: str(frozenset(x)))
    expect["consequents"] = expect["consequents"].apply(lambda x: str(frozenset(x)))
    expect.sort_values(columns_ordered, inplace=True)
    expect.reset_index(inplace=True, drop=True)

    assert res_df.equals(expect), res_df


def test_datatypes():
    res_df = association_rules(df_freq_items)
    for i in res_df["antecedents"]:
        assert isinstance(i, frozenset) is True

    for i in res_df["consequents"]:
        assert isinstance(i, frozenset) is True

    # cast itemset-containing dataframe to set and
    # check if association_rule converts it internally
    # back to frozensets
    df_freq_items_copy = df_freq_items.copy()
    df_freq_items_copy["itemsets"] = df_freq_items_copy["itemsets"].apply(
        lambda x: set(x)
    )

    res_df = association_rules(df_freq_items)
    for i in res_df["antecedents"]:
        assert isinstance(i, frozenset) is True

    for i in res_df["consequents"]:
        assert isinstance(i, frozenset) is True


def test_no_support_col():
    df_no_support_col = df_freq_items.loc[:, ["itemsets"]]
    numpy_assert_raises(ValueError, association_rules, df_no_support_col)


def test_no_itemsets_col():
    df_no_itemsets_col = df_freq_items.loc[:, ["support"]]
    numpy_assert_raises(ValueError, association_rules, df_no_itemsets_col)


def test_wrong_metric():
    numpy_assert_raises(ValueError, association_rules, df_freq_items, "unicorn")


def test_empty_result():
    expect = pd.DataFrame(
        columns=[
            "antecedents",
            "consequents",
            "antecedent support",
            "consequent support",
            "support",
            "confidence",
            "lift",
            "leverage",
            "conviction",
            "zhangs_metric",
        ]
    )
    res_df = association_rules(df_freq_items, min_threshold=2)
    assert res_df.equals(expect)


def test_leverage():
    res_df = association_rules(df_freq_items, min_threshold=0.1, metric="leverage")
    assert res_df.values.shape[0] == 6

    res_df = association_rules(
        df_freq_items_with_colnames, min_threshold=0.1, metric="leverage"
    )
    assert res_df.values.shape[0] == 6


def test_conviction():
    res_df = association_rules(df_freq_items, min_threshold=1.5, metric="conviction")
    assert res_df.values.shape[0] == 11

    res_df = association_rules(
        df_freq_items_with_colnames, min_threshold=1.5, metric="conviction"
    )
    assert res_df.values.shape[0] == 11


def test_lift():
    res_df = association_rules(df_freq_items, min_threshold=1.1, metric="lift")
    assert res_df.values.shape[0] == 6

    res_df = association_rules(
        df_freq_items_with_colnames, min_threshold=1.1, metric="lift"
    )
    assert res_df.values.shape[0] == 6


def test_confidence():
    res_df = association_rules(df_freq_items, min_threshold=0.8, metric="confidence")
    assert res_df.values.shape[0] == 9

    res_df = association_rules(
        df_freq_items_with_colnames, min_threshold=0.8, metric="confidence"
    )
    assert res_df.values.shape[0] == 9


def test_frozenset_selection():
    res_df = association_rules(df_freq_items)

    sel = res_df[res_df["consequents"] == frozenset((3, 5))]
    assert sel.values.shape[0] == 1

    sel = res_df[res_df["consequents"] == frozenset((5, 3))]
    assert sel.values.shape[0] == 1

    sel = res_df[res_df["consequents"] == {3, 5}]
    assert sel.values.shape[0] == 1

    sel = res_df[res_df["antecedents"] == frozenset((8, 3))]
    assert sel.values.shape[0] == 1


def test_override_metric_with_support():
    res_df = association_rules(df_freq_items_with_colnames, min_threshold=0.8)
    # default metric is confidence
    assert res_df.values.shape[0] == 9

    res_df = association_rules(
        df_freq_items_with_colnames, min_threshold=0.8, metric="support"
    )
    assert res_df.values.shape[0] == 2

    res_df = association_rules(
        df_freq_items_with_colnames, min_threshold=0.8, support_only=True
    )
    assert res_df.values.shape[0] == 2


def test_on_df_with_missing_entries():
    # this is a data frame where information about
    # antecedents and consequents have been cropped
    # see https://github.com/rasbt/mlxtend/issues/390
    # for more details
    dict = {
        "itemsets": [
            ["177", "176"],
            ["177", "179"],
            ["176", "178"],
            ["176", "179"],
            ["93", "100"],
            ["177", "178"],
            ["177", "176", "178"],
        ],
        "support": [
            0.253623,
            0.253623,
            0.217391,
            0.217391,
            0.181159,
            0.108696,
            0.108696,
        ],
    }

    df = pd.DataFrame(dict)

    numpy_assert_raises(KeyError, association_rules, df)


def test_on_df_with_missing_entries_support_only():
    # this is a data frame where information about
    # antecedents and consequents have been cropped
    # see https://github.com/rasbt/mlxtend/issues/390
    # for more details
    dict = {
        "itemsets": [
            ["177", "176"],
            ["177", "179"],
            ["176", "178"],
            ["176", "179"],
            ["93", "100"],
            ["177", "178"],
            ["177", "176", "178"],
        ],
        "support": [
            0.253623,
            0.253623,
            0.217391,
            0.217391,
            0.181159,
            0.108696,
            0.108696,
        ],
    }

    df = pd.DataFrame(dict)
    df_result = association_rules(df, support_only=True, min_threshold=0.1)

    assert df_result["support"].shape == (18,)
    assert int(np.isnan(df_result["support"].values).any()) != 1


def test_with_empty_dataframe():
    df = df_freq_items_with_colnames.iloc[:0]
    with pytest.raises(ValueError):
        association_rules(df)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Nonparametric Permutation Test
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from itertools import combinations, product
from math import factorial

import numpy as np

try:
    from nose.tools import nottest
except ImportError:
    # Use a no-op decorator if nose is not available
    def nottest(f):
        return f


# decorator to prevent nose to consider
# this as a unit test due to "test" in the name
@nottest
def permutation_test(
    x,
    y,
    func="x_mean != y_mean",
    method="exact",
    num_rounds=1000,
    seed=None,
    paired=False,
):
    """
    Nonparametric permutation test

    Parameters
    -------------
    x : list or numpy array with shape (n_datapoints,)
        A list or 1D numpy array of the first sample
        (e.g., the treatment group).

    y : list or numpy array with shape (n_datapoints,)
        A list or 1D numpy array of the second sample
        (e.g., the control group).

    func : custom function or str (default: 'x_mean != y_mean')
        function to compute the statistic for the permutation test.
        - If 'x_mean != y_mean', uses
          `func=lambda x, y: np.abs(np.mean(x) - np.mean(y)))`
           for a two-sided test.
        - If 'x_mean > y_mean', uses
          `func=lambda x, y: np.mean(x) - np.mean(y))`
           for a one-sided test.
        - If 'x_mean < y_mean', uses
          `func=lambda x, y: np.mean(y) - np.mean(x))`
           for a one-sided test.

    method : 'approximate' or 'exact' (default: 'exact')
        If 'exact' (default), all possible permutations are considered.
        If 'approximate' the number of drawn samples is
        given by `num_rounds`.
        Note that 'exact' is typically not feasible unless the dataset
        size is relatively small.

    paired : bool
        If True, a paired test is performed by only exchanging each
        datapoint with its associate.

    num_rounds : int (default: 1000)
        The number of permutation samples if `method='approximate'`.

    seed : int or None (default: None)
        The random seed for generating permutation samples if
        `method='approximate'`.

    Returns
    ----------
    p-value under the null hypothesis
    Examples

    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/permutation_test/

    """

    if method not in ("approximate", "exact"):
        raise AttributeError(
            'method must be "approximate"' ' or "exact", got %s' % method
        )

    if isinstance(func, str):
        if func not in ("x_mean != y_mean", "x_mean > y_mean", "x_mean < y_mean"):
            raise AttributeError(
                "Provide a custom function"
                " lambda x,y: ... or a string"
                ' in ("x_mean != y_mean", '
                '"x_mean > y_mean", "x_mean < y_mean")'
            )

        elif func == "x_mean != y_mean":

            def func(x, y):
                return np.abs(np.mean(x) - np.mean(y))

        elif func == "x_mean > y_mean":

            def func(x, y):
                return np.mean(x) - np.mean(y)

        else:

            def func(x, y):
                return np.mean(y) - np.mean(x)

    rng = np.random.RandomState(seed)

    m, n = len(x), len(y)

    if paired:
        if m != n:
            raise ValueError("x and y must have the same" " length if `paired=True`")
        sample_x = np.empty(m)
        sample_y = np.empty(n)

    else:
        combined = np.hstack((x, y))

    at_least_as_extreme = 0.0
    reference_stat = func(x, y)

    # Note that whether we compute the combinations or permutations
    # does not affect the results, since the number of permutations
    # n_A specific objects in A and n_B specific objects in B is the
    # same for all combinations in x_1, ... x_{n_A} and
    # x_{n_{A+1}}, ... x_{n_A + n_B}
    # In other words, for any given number of combinations, we get
    # n_A! x n_B! times as many permutations; hoewever, the computed
    # value of those permutations that are merely re-arranged combinations
    # does not change. Hence, the result, since we divide by the number of
    # combinations or permutations is the same, the permutations simply have
    # "n_A! x n_B!" as a scaling factor in the numerator and denominator
    # and using combinations instead of permutations simply saves computational
    # time

    if method == "exact":
        if paired:
            for flip in product([True, False], repeat=m):
                for i, f in enumerate(flip):
                    if f:
                        sample_x[i], sample_y[i] = y[i], x[i]
                    else:
                        sample_x[i], sample_y[i] = x[i], y[i]

                diff = func(sample_x, sample_y)
                if diff > reference_stat or np.isclose(diff, reference_stat):
                    at_least_as_extreme += 1.0

            num_rounds = 2**n

        else:
            for indices_x in combinations(range(m + n), m):
                indices_y = [i for i in range(m + n) if i not in indices_x]
                diff = func(combined[list(indices_x)], combined[indices_y])

                if diff > reference_stat or np.isclose(diff, reference_stat):
                    at_least_as_extreme += 1.0
            num_rounds = factorial(m + n) / (factorial(m) * factorial(n))

    else:
        if paired:
            for i in range(num_rounds):
                flip = rng.randn(m) > 0.0

                for i, f in enumerate(flip):
                    if f:
                        sample_x[i], sample_y[i] = y[i], x[i]
                    else:
                        sample_x[i], sample_y[i] = x[i], y[i]

                diff = func(sample_x, sample_y)
                if diff > reference_stat or np.isclose(diff, reference_stat):
                    at_least_as_extreme += 1.0

            # To cover the actual experiment results
            at_least_as_extreme += 1.0
            num_rounds += 1.0

        else:
            for i in range(num_rounds):
                rng.shuffle(combined)
                diff = func(combined[:m], combined[m:])

                if diff > reference_stat or np.isclose(diff, reference_stat):
                    at_least_as_extreme += 1.0

            # To cover the actual experiment results
            at_least_as_extreme += 1.0
            num_rounds += 1.0

    return at_least_as_extreme / num_rounds



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
from scipy.sparse import csr_matrix
from sklearn.base import clone

from mlxtend.preprocessing import TransactionEncoder
from mlxtend.utils import assert_raises

dataset = [
    ["Apple", "Beer", "Rice", "Chicken"],
    ["Apple", "Beer", "Rice"],
    ["Apple", "Beer"],
    ["Apple", "Bananas"],
    ["Milk", "Beer", "Rice", "Chicken"],
    ["Milk", "Beer", "Rice"],
    ["Milk", "Beer"],
    ["Apple", "Bananas"],
]

data_sorted = [
    ["Apple", "Beer", "Chicken", "Rice"],
    ["Apple", "Beer", "Rice"],
    ["Apple", "Beer"],
    ["Apple", "Bananas"],
    ["Beer", "Chicken", "Milk", "Rice"],
    ["Beer", "Milk", "Rice"],
    ["Beer", "Milk"],
    ["Apple", "Bananas"],
]


expect = np.array(
    [
        [1, 0, 1, 1, 0, 1],
        [1, 0, 1, 0, 0, 1],
        [1, 0, 1, 0, 0, 0],
        [1, 1, 0, 0, 0, 0],
        [0, 0, 1, 1, 1, 1],
        [0, 0, 1, 0, 1, 1],
        [0, 0, 1, 0, 1, 0],
        [1, 1, 0, 0, 0, 0],
    ]
)


def test_fit():
    oht = TransactionEncoder()
    oht.fit(dataset)
    assert oht.columns_ == ["Apple", "Bananas", "Beer", "Chicken", "Milk", "Rice"]


def test_transform():
    oht = TransactionEncoder()
    oht.fit(dataset)
    trans = oht.transform(dataset)
    np.testing.assert_array_equal(expect, trans)


def test_transform_sparse():
    oht = TransactionEncoder()
    oht.fit(dataset)
    trans = oht.transform(dataset, sparse=True)
    assert isinstance(trans, csr_matrix)
    np.testing.assert_array_equal(expect, trans.todense())


def test_fit_transform():
    oht = TransactionEncoder()
    trans = oht.fit_transform(dataset)
    np.testing.assert_array_equal(expect, trans)


def test_inverse_transform():
    oht = TransactionEncoder()
    oht.fit(dataset)
    np.testing.assert_array_equal(
        np.array(data_sorted), np.array(oht.inverse_transform(expect))
    )


def test_cloning():
    oht = TransactionEncoder()
    oht.fit(dataset)
    oht2 = clone(oht)

    msg = "'TransactionEncoder' object has no attribute 'columns_'"
    assert_raises(AttributeError, msg, oht2.transform, dataset)

    trans = oht2.fit_transform(dataset)
    np.testing.assert_array_equal(expect, trans)



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from sklearn.datasets import make_moons

from mlxtend.feature_extraction import RBFKernelPCA as KPCA

X1, y1 = make_moons(n_samples=50, random_state=1)


def test_default_components():
    pca = KPCA()
    pca.fit(X1)
    assert pca.X_projected_.shape == X1.shape


def test_default_2components():
    pca = KPCA(n_components=2)
    pca.fit(X1)
    assert pca.X_projected_.shape == (X1.shape[0], 2)


def test_default_0components():
    with pytest.raises(AttributeError):
        pca = KPCA(n_components=0)
        pca.fit(X1)


def test_proj():
    pca = KPCA(n_components=2)
    pca.fit(X1[:2])
    exp = np.array([[-0.71, -0.71], [0.71, -0.71]])
    assert_almost_equal(pca.X_projected_, exp, decimal=2)


def test_reproj_1():
    pca = KPCA(n_components=2)
    pca.fit(X1)
    exp = pca.transform(X1)
    assert_almost_equal(pca.X_projected_, exp, decimal=2)


def test_reproj_2():
    pca = KPCA(n_components=2)
    pca.fit(X1)
    exp = pca.transform(X1[1, None])
    assert_almost_equal(pca.X_projected_[1, None], exp, decimal=2)


def test_fail_array_fit():
    pca = KPCA(n_components=2)
    with pytest.raises(ValueError):
        pca.fit(X1[1])


def test_fail_array_transform():
    pca = KPCA(n_components=2)
    pca.fit(X1)
    with pytest.raises(ValueError):
        pca.transform(X1[1])



# mlxtend Machine Learning Library Extensions
# Author: Steve Harenberg <harenbergsd@gmail.com>
#
# License: BSD 3 clause

import itertools
import math

from ..frequent_patterns import fpcommon as fpc


def fpgrowth(df, min_support=0.5, use_colnames=False, max_len=None, verbose=0):
    """Get frequent itemsets from a one-hot DataFrame

    Parameters
    -----------
    df : pandas DataFrame
      pandas DataFrame the encoded format. Also supports
      DataFrames with sparse data; for more info, please
      see https://pandas.pydata.org/pandas-docs/stable/user_guide/sparse.html#sparse-data-structures.

      Please note that the old pandas SparseDataFrame format
      is no longer supported in mlxtend >= 0.17.2.

      The allowed values are either 0/1 or True/False.
      For example,

    ```
           Apple  Bananas   Beer  Chicken   Milk   Rice
        0   True    False   True     True  False   True
        1   True    False   True    False  False   True
        2   True    False   True    False  False  False
        3   True     True  False    False  False  False
        4  False    False   True     True   True   True
        5  False    False   True    False   True   True
        6  False    False   True    False   True  False
        7   True     True  False    False  False  False
    ```

    min_support : float (default: 0.5)
      A float between 0 and 1 for minimum support of the itemsets returned.
      The support is computed as the fraction
      transactions_where_item(s)_occur / total_transactions.

    use_colnames : bool (default: False)
      If true, uses the DataFrames' column names in the returned DataFrame
      instead of column indices.

    max_len : int (default: None)
      Maximum length of the itemsets generated. If `None` (default) all
      possible itemsets lengths are evaluated.

    verbose : int (default: 0)
      Shows the stages of conditional tree generation.

    Returns
    -----------
    pandas DataFrame with columns ['support', 'itemsets'] of all itemsets
      that are >= `min_support` and < than `max_len`
      (if `max_len` is not None).
      Each itemset in the 'itemsets' column is of type `frozenset`,
      which is a Python built-in type that behaves similarly to
      sets except that it is immutable
      (For more info, see
      https://docs.python.org/3.6/library/stdtypes.html#frozenset).

    Examples
    ----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/frequent_patterns/fpgrowth/

    """
    fpc.valid_input_check(df)

    if min_support <= 0.0:
        raise ValueError(
            "`min_support` must be a positive "
            "number within the interval `(0, 1]`. "
            "Got %s." % min_support
        )

    colname_map = None
    if use_colnames:
        colname_map = {idx: item for idx, item in enumerate(df.columns)}

    tree, _ = fpc.setup_fptree(df, min_support)
    minsup = math.ceil(min_support * len(df.index))  # min support as count
    generator = fpg_step(tree, minsup, colname_map, max_len, verbose)

    return fpc.generate_itemsets(generator, len(df.index), colname_map)


def fpg_step(tree, minsup, colnames, max_len, verbose):
    """
    Performs a recursive step of the fpgrowth algorithm.

    Parameters
    ----------
    tree : FPTree
    minsup : int

    Yields
    ------
    lists of strings
        Set of items that has occurred in minsup itemsets.
    """
    count = 0
    items = tree.nodes.keys()
    if tree.is_path():
        # If the tree is a path, we can combinatorally generate all
        # remaining itemsets without generating additional conditional trees
        size_remain = len(items) + 1
        if max_len:
            size_remain = max_len - len(tree.cond_items) + 1
        for i in range(1, size_remain):
            for itemset in itertools.combinations(items, i):
                count += 1
                support = min([tree.nodes[i][0].count for i in itemset])
                yield support, tree.cond_items + list(itemset)
    elif not max_len or max_len > len(tree.cond_items):
        for item in items:
            count += 1
            support = sum([node.count for node in tree.nodes[item]])
            yield support, tree.cond_items + [item]

    if verbose:
        tree.print_status(count, colnames)

    # Generate conditional trees to generate frequent itemsets one item larger
    if not tree.is_path() and (not max_len or max_len > len(tree.cond_items)):
        for item in items:
            cond_tree = tree.conditional_tree(item, minsup)
            for sup, iset in fpg_step(cond_tree, minsup, colnames, max_len, verbose):
                yield sup, iset



import pytest
from packaging.version import Version
from sklearn import __version__ as sklearn_version

from mlxtend.data import iris_data
from mlxtend.plotting import plot_pca_correlation_graph

if Version(sklearn_version) < "0.22":
    from sklearn.decomposition.pca import PCA
else:
    from sklearn.decomposition import PCA


def test_pass_pca_corr():
    X, y = iris_data()
    plot_pca_correlation_graph(X, variables_names=["1", "2", "3", "4"])


def test_pass_pca_corr_pca_out():
    X, y = iris_data()
    pca = PCA(n_components=2)
    X_pca = pca.fit_transform(X)
    eigen = pca.explained_variance_

    plot_pca_correlation_graph(
        X, variables_names=["1", "2", "3", "4"], X_pca=X_pca, explained_variance=eigen
    )


def test_X_PCA_but_no_explained_variance():
    with pytest.raises(
        ValueError,
        match="If `X_pca` is not None, the `explained variance` "
        "values should not be `None`.",
    ):
        X, y = iris_data()
        pca = PCA(n_components=2)
        X_pca = pca.fit_transform(X)

        plot_pca_correlation_graph(
            X,
            variables_names=["1", "2", "3", "4"],
            X_pca=X_pca,
            explained_variance=None,
        )


def test_no_X_PCA_but_explained_variance():
    with pytest.raises(
        ValueError,
        match="If `explained variance` is not None, the "
        "`X_pca` values should not be `None`.",
    ):
        X, y = iris_data()
        pca = PCA(n_components=2)
        pca.fit(X)
        eigen = pca.explained_variance_

        plot_pca_correlation_graph(
            X,
            variables_names=["1", "2", "3", "4"],
            X_pca=None,
            explained_variance=eigen,
        )


def test_not_enough_components():
    s = (
        "Number of principal components must match the number"
        " of eigenvalues. Got 2 != 1"
    )
    with pytest.raises(ValueError, match=s):
        X, y = iris_data()
        pca = PCA(n_components=2)
        X_pca = pca.fit_transform(X)
        eigen = pca.explained_variance_

        plot_pca_correlation_graph(
            X,
            variables_names=["1", "2", "3", "4"],
            X_pca=X_pca,
            explained_variance=eigen[:-1],
        )
